# -*- coding=utf -*-

import json
import logging
from ..logging import get_logger
from ..browser import *

class SlicerBrowser(AggregationBrowser):
    """Aggregation browser for Cubes Slicer OLAP server."""

    def __init__(self, cube, store, locale=None, **options):
        """Browser for another Slicer server.
        """
        super(SlicerBrowser, self).__init__(cube, store, locale)

        self.logger = get_logger()
        self.cube = cube
        self.locale = locale
        self.store = store

    def features(self):

        # Get the original features as provided by the Slicer server.
        # They are stored in browser_options in the Slicer model provider's
        # cube().
        features = dict(self.cube.browser_options.get("features", {}))

        # Replace only the actions, as we are not just a simple proxy.
        features["actions"] = ["aggregate", "facts", "fact", "cell", "members"]

        return features

    def provide_aggregate(self, cell, aggregates, drilldown, split, order,
                          page, page_size, **options):

        params = {}

        if cell:
            params["cut"] = string_from_cuts(cell.cuts)

        if drilldown:
            params["drilldown"] = ",".join(drilldown.items_as_strings())

        if split:
            params["split"] = str(split)

        if aggregates:
            names = [a.name for a in aggregates]
            params["aggregates"] = ",".join(names)

        if order:
            params["order"] = self._order_param(order)

        if page is not None:
            params["page"] = str(page)

        if page_size is not None:
            params["page_size"] = str(page_size)


        response = self.store.cube_request("aggregate",
                                           self.cube.basename, params)

        result = AggregationResult()

        result.cells = response.get('cells', [])

        if "summary" in response:
            result.summary = response.get('summary')

        result.levels = response.get('levels', {})
        result.labels = response.get('labels', [])
        result.cell = cell
        result.aggregates = response.get('aggregates', [])

        return result

    def facts(self, cell=None, fields=None, order=None, page=None,
              page_size=None):

        cell = cell or Cell(self.cube)
        if fields:
            attributes = self.cube.get_attributes(fields)
        else:
            attributes = []

        order = self.prepare_order(order, is_aggregate=False)

        params = {}

        if cell:
            params["cut"] = string_from_cuts(cell.cuts)

        if order:
            params["order"] = self._order_param(order)

        if page is not None:
            params["page"] = str(page)

        if page_size is not None:
            params["page_size"] = str(page_size)

        if attributes:
            params["fields"] = ",".join(str(attr) for attr in attributes)

        params["format"] = "json_lines"

        response = self.store.cube_request("facts", self.cube.basename, params,
                                           is_lines=True)

        return Facts(response, attributes)

    def provide_members(self, cell=None, dimension=None, levels=None,
                        hierarchy=None, attributes=None, page=None,
                        page_size=None, order=None, **options):

        params = {}

        if cell:
            params["cut"] = string_from_cuts(cell.cuts)

        if order:
            params["order"] = self._order_param(order)

        if levels:
            params["level"] = str(levels[-1])

        if hierarchy:
            params["hierarchy"] = str(hierarchy)

        if page is not None:
            params["page"] = str(page)

        if page_size is not None:
            params["page_size"] = str(page_size)

        if attributes:
            params["fields"] = ",".join(str(attr) for attr in attributes)

        params["format"] = "json_lines"

        action = "/cube/%s/members/%s" % (self.cube.basename, str(dimension))
        response = self.store.request(action, params, is_lines=True)

        return response

    def cell_details(self, cell, dimension=None):
        cell = cell or Cell(self.cube)

        params = {}
        if cell:
            params["cut"] = string_from_cuts(cell.cuts)

        if dimension:
            params["dimension"] = str(dimension)

        response = self.store.cube_request("cell", self.cube.basename, params) 

        return response

    def fact(self, fact_id):
        action = "/cube/%s/fact/%s" % (self.cube.basename, str(fact_id))
        response = self.store.request(action)
        return response

    def is_builtin_function(self, name, aggregate):
        return True

    def _order_param(self, order):
        """Prepare an order string in form: ``attribute:direction``"""
        string = ",".join("%s:%s" % (o[0], o[1]) for o in order)
        return string




# -*- coding=utf -*-

from __future__ import absolute_import

from ...server.logging import RequestLogHandler, REQUEST_LOG_ITEMS
from sqlalchemy import create_engine, Table, MetaData, Column
from sqlalchemy import Integer, Sequence, DateTime, String, Float
from sqlalchemy.exc import NoSuchTableError
from ...browser import Drilldown
from .store import create_sqlalchemy_engine

import logging

class SQLRequestLogHandler(RequestLogHandler):
    def __init__(self, url=None, table=None, dimensions_table=None, **options):

        self.url = url
        self.engine = create_sqlalchemy_engine(url, options)

        metadata = MetaData(bind=self.engine)

        logging.getLogger('sqlalchemy.engine').setLevel("DEBUG")
        logging.getLogger('sqlalchemy.pool').setLevel("DEBUG")

        try:
            self.table = Table(table, metadata, autoload=True)

        except NoSuchTableError:
            columns = [
                Column('id', Integer, Sequence(table+"_seq"),
                       primary_key=True),
                Column('timestamp', DateTime),
                Column('method', String(50)),
                Column('cube', String(250)),
                Column('cell', String(2000)),
                Column('identity', String(250)),
                Column('elapsed_time', Float),
                Column('attributes', String(2000)),
                Column('split', String(2000)),
                Column('drilldown', String(2000)),
                Column('page', Integer),
                Column('page_size', Integer),
                Column('format', String(50)),
                Column('header', String(50)),
            ]

            self.table = Table(table, metadata, extend_existing=True, *columns)
            self.table.create()

        # Dimensions table: use of dimensions
        # Used-as: cut, split, drilldown
        # Value:
        #     cut: cut value
        #     split: cut value

        if dimensions_table:
            try:
                self.dims_table = Table(dimensions_table, metadata, autoload=True)

            except NoSuchTableError:
                columns = [
                    Column('id', Integer, Sequence(table+"_seq"),
                           primary_key=True),
                    Column('query_id', Integer),
                    Column('dimension', String(250)),
                    Column('hierarchy', String(250)),
                    Column('level', String(250)),
                    Column('used_as', String(50)),
                    Column('value', String(2000)),
                ]

                self.dims_table = Table(dimensions_table, metadata, extend_existing=True, *columns)
                self.dims_table.create()
        else:
            self.dims_table = None

    def write_record(self, cube, cell, record):
        drilldown = record.get("drilldown")

        if drilldown is not None:
            if cell:
                drilldown = Drilldown(drilldown, cell)
                record["drilldown"] = str(drilldown)
            else:
                drilldown = []
                record["drilldown"] = None

        connection = self.engine.connect()
        trans = connection.begin()

        insert = self.table.insert().values(record)
        result = connection.execute(insert)
        query_id = result.inserted_primary_key[0]

        if self.dims_table is not None:
            uses = []

            cuts = cell.cuts if cell else []
            cuts = cuts or []

            for cut in cuts:
                dim = cube.dimension(cut.dimension)
                depth = cut.level_depth()
                if depth:
                    level = dim.hierarchy(cut.hierarchy)[depth-1]
                    level_name = str(level)
                else:
                    level_name = None

                use = {
                    "query_id": query_id,
                    "dimension": str(dim),
                    "hierarchy": str(cut.hierarchy),
                    "level": str(level_name),
                    "used_as": "cell",
                    "value": str(cut)
                }
                uses.append(use)

            if drilldown:
                for item in drilldown:
                    (dim, hier, levels) = item[0:3]
                    if levels:
                        level = str(levels[-1])
                    else:
                        level = None

                    use = {
                        "query_id": query_id,
                        "dimension": str(dim),
                        "hierarchy": str(hier),
                        "level": str(level),
                        "used_as": "drilldown",
                        "value": None
                    }
                    uses.append(use)


            if uses:
                insert = self.dims_table.insert().values(uses)
                connection.execute(insert)

        trans.commit()
        connection.close()




# -*- encoding: utf-8 -*-
"""Cubes SQL backend utilities, mostly to be used by the slicer command."""

from sqlalchemy.sql.expression import Executable, ClauseElement
from sqlalchemy.ext.compiler import compiles
import sqlalchemy.sql as sql

from collections import OrderedDict

from ..browser import SPLIT_DIMENSION_NAME

__all__ = [
    "CreateTableAsSelect",
    "InsertIntoAsSelect",
    "CreateOrReplaceView",
    "condition_conjunction",
    "order_column",
    "order_query",
    "paginate_query"
]

class CreateTableAsSelect(Executable, ClauseElement):
    def __init__(self, table, select):
        self.table = table
        self.select = select

@compiles(CreateTableAsSelect)
def visit_create_table_as_select(element, compiler, **kw):
    preparer = compiler.dialect.preparer(compiler.dialect)
    full_name = preparer.format_table(element.table)

    return "CREATE TABLE %s AS (%s)" % (
        element.table,
        compiler.process(element.select)
    )
@compiles(CreateTableAsSelect, "sqlite")
def visit_create_table_as_select(element, compiler, **kw):
    preparer = compiler.dialect.preparer(compiler.dialect)
    full_name = preparer.format_table(element.table)

    return "CREATE TABLE %s AS %s" % (
        element.table,
        compiler.process(element.select)
    )

class CreateOrReplaceView(Executable, ClauseElement):
    def __init__(self, view, select):
        self.view = view
        self.select = select

@compiles(CreateOrReplaceView)
def visit_create_or_replace_view(element, compiler, **kw):
    preparer = compiler.dialect.preparer(compiler.dialect)
    full_name = preparer.format_table(element.view)

    return "CREATE OR REPLACE VIEW %s AS (%s)" % (
        full_name,
        compiler.process(element.select)
    )

@compiles(CreateOrReplaceView, "sqlite")
def visit_create_or_replace_view(element, compiler, **kw):
    preparer = compiler.dialect.preparer(compiler.dialect)
    full_name = preparer.format_table(element.view)

    return "CREATE VIEW %s AS %s" % (
        full_name,
        compiler.process(element.select)
    )

@compiles(CreateOrReplaceView, "mysql")
def visit_create_or_replace_view(element, compiler, **kw):
    preparer = compiler.dialect.preparer(compiler.dialect)
    full_name = preparer.format_table(element.view)

    return "CREATE OR REPLACE VIEW %s AS %s" % (
        full_name,
        compiler.process(element.select)
    )


def paginate_query(statement, page, page_size):
    """Returns paginated statement if page is provided, otherwise returns
    the same statement."""

    if page is not None and page_size is not None:
        statement = statement.offset(page * page_size).limit(page_size)

    return statement


def order_column(column, order):
    """Orders a `column` according to `order` specified as string. Returns a
    `Column` expression"""

    if not order:
        return column
    elif order.lower().startswith("asc"):
        return column.asc()
    elif order.lower().startswith("desc"):
        return column.desc()
    else:
        raise ArgumentError("Unknown order %s for column %s") % (order, column)


def order_query(statement, order, natural_order=None, labels=None):
    """Returns a SQL statement which is ordered according to the `order`. If
    the statement contains attributes that have natural order specified, then
    the natural order is used, if not overriden in the `order`.

    * `statement` – statement to be ordered
    * `order` explicit order, list of tuples (`aggregate`, `direction`)
    * `natural_order` – natural order of attributes in the statement – a
       dictionary where keys are attribute names and vales are directions.
       Used to look-up the natural order.
    * `labels` – mapping between logical labels and physical labels. Important
      when `safe_labels` is enabled. Read more about `safe_labels` for more
      information.
    """

    order = order or []
    labels = labels or {}
    natural_order = natural_order or []

    final_order = OrderedDict()

    # Each attribute mentioned in the order should be present in the selection
    # or as some column from joined table. Here we get the list of already
    # selected columns and derived aggregates

    # Get logical attributes from column labels (see logical_labels
    # description for more information why this step is necessary)

    columns = OrderedDict(zip(labels, statement.columns))

    # Normalize order
    # ---------------
    # Make sure that the `order` is a list of of tuples (`attribute`,
    # `order`). If element of the `order` list is a string, then it is
    # converted to (`string`, ``None``).

    if SPLIT_DIMENSION_NAME in statement.columns:
        split_column = sql.expression.column(SPLIT_DIMENSION_NAME)
        final_order[SPLIT_DIMENSION_NAME] = split_column

    # Collect the corresponding attribute columns
    for attribute, direction in order:
        attribute = str(attribute)
        column = order_column(columns[attribute], direction)

        if attribute not in final_order:
            final_order[attribute] = column

    # Collect natural order for selected columns that have no explicit
    # ordering
    for (name, column) in columns.items():
        if name in natural_order and name not in order_by:
            final_order[name] = order_column(column, natural_order[name])

    statement = statement.order_by(*final_order.values())

    return statement




# -*- coding: utf-8 -*-
from flask import Blueprint, Flask, Response, request, g, current_app
from functools import wraps

from ..workspace import Workspace
from ..auth import NotAuthorized
from ..cells import Cell, cut_from_dict
from ..browser import SPLIT_DIMENSION_NAME
from ..cells import cuts_from_string
from ..errors import *
from .utils import *
from .errors import *
from .local import *
from ..calendar import CalendarMemberConverter

from contextlib import contextmanager

# Utils
# -----

def prepare_cell(argname="cut", target="cell", restrict=False):
    """Sets `g.cell` with a `Cell` object from argument with name `argname`"""
    # Used by prepare_browser_request and in /aggregate for the split cell


    # TODO: experimental code, for now only for dims with time role
    converters = {
        "time": CalendarMemberConverter(workspace.calendar)
    }

    cuts = []
    for cut_string in request.args.getlist(argname):
        cuts += cuts_from_string(g.cube, cut_string,
                                 role_member_converters=converters)

    if cuts:
        cell = Cell(g.cube, cuts)
    else:
        cell = None

    if restrict:
        if workspace.authorizer:
            cell = workspace.authorizer.restricted_cell(g.auth_identity,
                                                        cube=g.cube,
                                                        cell=cell)
    setattr(g, target, cell)


def requires_cube(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if "lang" in request.args:
            g.locale = request.args.get("lang")
        else:
            g.locale = None

        cube_name = request.view_args.get("cube_name")
        try:
            g.cube = authorized_cube(cube_name, locale=g.locale)
        except NoSuchCubeError:
            raise NotFoundError(cube_name, "cube",
                                "Unknown cube '%s'" % cube_name)

        return f(*args, **kwargs)

    return wrapper

def requires_browser(f):
    """Prepares three global variables: `g.cube`, `g.browser` and `g.cell`.
    Also athorizes the cube using `authorize()`."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        if "lang" in request.args:
            g.locale = request.args.get("lang")
        else:
            g.locale = None

        cube_name = request.view_args.get("cube_name")
        if cube_name:
            cube = authorized_cube(cube_name, g.locale)
        else:
            cube = None

        g.cube = cube
        g.browser = workspace.browser(g.cube)

        prepare_cell(restrict=True)

        if "page" in request.args:
            try:
                g.page = int(request.args.get("page"))
            except ValueError:
                raise RequestError("'page' should be a number")
        else:
            g.page = None

        if "pagesize" in request.args:
            try:
                g.page_size = int(request.args.get("pagesize"))
            except ValueError:
                raise RequestError("'pagesize' should be a number")
        else:
            g.page_size = None

        # Collect orderings:
        # order is specified as order=<field>[:<direction>]
        #
        g.order = []
        for orders in request.args.getlist("order"):
            for order in orders.split(","):
                split = order.split(":")
                if len(split) == 1:
                    g.order.append( (order, None) )
                else:
                    g.order.append( (split[0], split[1]) )

        return f(*args, **kwargs)

    return wrapper


# Get authorized cube
# ===================

def authorized_cube(cube_name, locale):
    """Returns a cube `cube_name`. Handle cube authorization if required."""

    try:
        cube = workspace.cube(cube_name, g.auth_identity, locale=locale)
    except NotAuthorized:
        ident = "'%s'" % g.auth_identity if g.auth_identity \
                        else "unspecified identity"
        raise NotAuthorizedError("Authorization of cube '%s' failed for "
                                 "%s" % (cube_name, ident))
    return cube


# Query Logging
# =============

def log_request(action, attrib_field="attributes"):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            rlogger = current_app.slicer.request_logger

            # TODO: move this to request wrapper (same code as in aggregate)
            ddlist = request.args.getlist("drilldown")
            drilldown = []
            if ddlist:
                for ddstring in ddlist:
                    drilldown += ddstring.split("|")

            other = {
                "split": request.args.get("split"),
                "drilldown": drilldown,
                "page": g.page,
                "page_size": g.page_size,
                "format": request.args.get("format"),
                "header": request.args.get("header"),
                "attributes": request.args.get(attrib_field)
            }

            with rlogger.log_time(action, g.browser, g.cell, g.auth_identity,
                                  **other):
                retval = f(*args, **kwargs)

            return retval

        return wrapper

    return decorator




# -*- coding: utf-8 -*-

import sqlalchemy
import csv
import codecs
from cubes import compat


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        assert 'b' in f.mode, "in py3k, codec's StreamReader needs a bytestream"
        self.reader = codecs.getreader(encoding)(f)
        self.next = self.__next__
    def __iter__(self):
        return self

    def __next__(self):
        return compat.to_unicode(next(self.reader))


class UnicodeReader:
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)
        self.next = self.__next__

    def __next__(self):
        row = next(self.reader)
        return [compat.to_unicode(s) for s in row]

    def __iter__(self):
        return self


def create_table_from_csv(connectable, file_name, table_name, fields,
                          create_id=False, schema=None):
    """Create a table with name `table_name` from a CSV file `file_name` with columns corresponding
    to `fields`. The `fields` is a list of two string tuples: (name, type) where type might be:
    ``integer``, ``float`` or ``string``.

    If `create_id` is ``True`` then a column with name ``id`` is created and will contain generated
    sequential record id.

    This is just small utility function for sandbox, play-around and testing purposes. It is not
    recommended to be used for serious CSV-to-table loadings. For more advanced CSV loadings use another
    framework, such as Brewery (http://databrewery.org).
    """

    metadata = sqlalchemy.MetaData(bind=connectable)

    table = sqlalchemy.Table(table_name, metadata, autoload=False, schema=schema)
    if table.exists():
        table.drop(checkfirst=False)

    type_map = {"integer": sqlalchemy.Integer,
                "float": sqlalchemy.Numeric,
                "string": sqlalchemy.String(256),
                "text": sqlalchemy.Text,
                "date": sqlalchemy.Text,
                "boolean": sqlalchemy.Integer}

    if create_id:
        col = sqlalchemy.schema.Column('id', sqlalchemy.Integer, primary_key=True)
        table.append_column(col)

    field_names = []
    for (field_name, field_type) in fields:
        col = sqlalchemy.schema.Column(field_name, type_map[field_type.lower()])
        table.append_column(col)
        field_names.append(field_name)

    table.create()

    reader = UnicodeReader(open(file_name, 'rb'))

    # Skip header
    next(reader)

    insert_command = table.insert()

    for row in reader:
        record = dict(zip(field_names, row))
        insert_command.execute(record)
